import json
import os
import pathlib
import tempfile
from typing import List
from unittest.mock import patch, Mock

from click.testing import CliRunner, Result
from docker.types import Mount
from pytest_mock import MockFixture

from odahuflow.cli.parsers.local import training
from odahuflow.sdk.gppi.executor import PROJECT_FILE
from odahuflow.sdk.local import training as training_sdk
from odahuflow.sdk.local.training import MODEL_OUTPUT_CONTAINER_PATH
from odahuflow.sdk.models import ModelTraining, ToolchainIntegration, ModelTrainingSpec


def test_list_local_trainings(tmpdir):
    folders = ['wine-1@-12',
               '&wine-1-12',
               '[wine-1@-12',
               '2Wine-1@-12',
               'Wine-1@-123',
               '@wine-1@-12',
               'zine-1@-12',
               'Awine-1@']
    for folder in folders:
        tmpdir.mkdir(folder)
        pathlib.Path(tmpdir, folder, PROJECT_FILE).touch()
    with patch('odahuflow.sdk.local.training.config.LOCAL_MODEL_OUTPUT_DIR', tmpdir):
        assert training_sdk.list_local_trainings() == ['&wine-1-12',
                                          '2Wine-1@-12',
                                          '@wine-1@-12',
                                          'Awine-1@',
                                          'Wine-1@-123',
                                          '[wine-1@-12',
                                          'wine-1@-12',
                                          'zine-1@-12']


def test_local_training_relative_output_dir(mocker: MockFixture, cli_runner: CliRunner):
    """
    Tests issue #208 - Converting relative path to output-dir to absolute,
    because Docker requires mount paths to be absolute
    :param mocker: mocker fixture
    :param cli_runner: Click runner fixture
    """

    trainer_mock = mocker.patch.object(training, 'K8sTrainer', autospec=True).return_value
    trainer_mock.model_training.spec.model.artifact_name_template = 'model_dir_template'
    api_client = Mock()

    mocker.patch.object(training_sdk, 'create_mt_config_file')
    mocker.patch.object(training_sdk, 'stream_container_logs')
    mocker.patch.object(training_sdk, 'raise_error_if_container_failed')

    docker_mock: Mock = mocker.patch.object(training_sdk.docker, 'from_env').return_value
    docker_mock.api.inspect_container = Mock(return_value={})

    with tempfile.TemporaryDirectory(dir=os.curdir) as temp_dir:
        temp_dir_path = pathlib.Path(temp_dir)
        training_yml = temp_dir_path / 'training.yml'
        training_yml.write_text(json.dumps(
            {**ModelTraining(id='training1', spec=ModelTrainingSpec(toolchain='toolchain1')).to_dict(),
             **{'kind': 'ModelTraining'}}))

        toolchain_yml = temp_dir_path / 'toolchain.yml'
        toolchain_yml.write_text(json.dumps({**ToolchainIntegration(id='toolchain1').to_dict(),
                                             **{'kind': 'ToolchainIntegration'}}))

        temp_dir_relative_path: str = os.path.relpath(temp_dir_path)

        result: Result = cli_runner.invoke(
            training.training_group,
            ['run', '--output-dir', temp_dir_relative_path, '--manifest-file', str(training_yml),
             '--manifest-file', str(toolchain_yml), '--id', 'training1'],
            obj=api_client)

        assert result.exit_code == 0, f'command invocation ended with exit code {result.exit_code}'
        assert result.exception is None, f'command invocation ended with exception {result.exception}'

        abs_path = os.path.abspath(temp_dir)

        for call in docker_mock.containers.run.call_args_list:
            call_kwargs = call[1]
            mounts: List[Mount] = call_kwargs.get('mounts', [])

            for mount in filter(lambda m: m.get('Target') == MODEL_OUTPUT_CONTAINER_PATH, mounts):
                assert os.path.dirname(mount.get('Source')) == abs_path



#  Copyright 2020 EPAM Systems
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import click


ABBREVIATION = {
    "conn": "connection",
    "conf": "config",
    "dep": "deployment",
    "pack": "packaging",
    "pi": "packaging-integration",
    "temp": "template",
    "ti": "toolchain-integration",
    "train": "training"
}


class BetterHelpGroup(click.Group):
    """
    This class allows user to get a subcommand's help text without invocation of group's callback

    This is to avoid bad user experience when user wants to check a subcommand's help text, but CLI asks him to
    fulfill all required arguments of a group. Also even with all required arguments in place, group's callback
    can fail for any reason, which is non-sense when asking CLI for a subcommand's help.
    """

    def invoke(self, ctx: click.Context):
        """
        This method sets a callback function (the one that decorated by click.group()) to None,
        if --help appears in args. So the callback is ignored at all if user asks to print help text.
        """
        if any(arg in ctx.help_option_names for arg in ctx.args):
            self.callback = None
        return super().invoke(ctx)


class AbbreviationGroup(BetterHelpGroup):
    """
    AbbreviationGroup
    """

    def get_command(self, ctx, cmd_name):
        """
        Override get command of click.Group
        :param ctx: click context
        :param cmd_name: group name
        :return: click Command
        """
        rv = click.Group.get_command(self, ctx, cmd_name)
        if rv is not None:
            return rv

        return click.Group.get_command(self, ctx, ABBREVIATION.get(cmd_name))



import base64
from typing import Any, Dict, List, Tuple
from unittest.mock import Mock

from dataclasses import dataclass
import pytest
import yaml
from click.testing import CliRunner
from odahuflow.cli.parsers.local.packaging import run
from odahuflow.sdk.clients.api import WrongHttpStatusCode
from odahuflow.sdk.local import packaging as packaging_sdk
from odahuflow.sdk.models import ConnectionSpec, K8sPackager, ModelPackaging, ModelPackagingSpec, ModelPackagingStatus, \
    PackagingIntegration, \
    PackagerTarget, \
    PackagingIntegrationSpec, Schema, Target, \
    Connection, TargetSchema
from pytest_mock import MockFixture

PLAIN_VALUE = 'Value'

conn1 = Connection(
    id='conn1',
    spec=ConnectionSpec(
        key_id=base64.b64encode(bytes(PLAIN_VALUE, 'utf-8')).decode('utf-8'),
        key_secret=base64.b64encode(bytes(PLAIN_VALUE, 'utf-8')).decode('utf-8'),
        public_key=base64.b64encode(bytes(PLAIN_VALUE, 'utf-8')).decode('utf-8'),
        password=base64.b64encode(bytes(PLAIN_VALUE, 'utf-8')).decode('utf-8'),
        type='gcr'
    )
)
conn2 = Connection(
    id='conn2',
    spec=ConnectionSpec(
        key_id=base64.standard_b64encode(bytes(PLAIN_VALUE, 'utf-8')).decode('utf-8'),
        key_secret=base64.b64encode(bytes(PLAIN_VALUE, 'utf-8')).decode('utf-8'),
        public_key=base64.b64encode(bytes(PLAIN_VALUE, 'utf-8')).decode('utf-8'),
        password=base64.b64encode(bytes(PLAIN_VALUE, 'utf-8')).decode('utf-8'),
        type='gcr'
    )
)
conn_d1 = Connection(
    id='conn1',
    spec=ConnectionSpec(
        key_id=PLAIN_VALUE,
        key_secret=PLAIN_VALUE,
        public_key=PLAIN_VALUE,
        password=PLAIN_VALUE,
        type='gcr'
    )
)
conn_d2 = Connection(
    id='conn2',
    spec=ConnectionSpec(
        key_id=PLAIN_VALUE,
        key_secret=PLAIN_VALUE,
        public_key=PLAIN_VALUE,
        password=PLAIN_VALUE,
        type='gcr'
    )
)
target_pull = Target("conn1", "docker-pull")
target_push = Target("conn2", "docker-push")
pi_target_pull = PackagerTarget(conn_d1, "docker-pull")
pi_target_push = PackagerTarget(conn_d2, "docker-push")


pack1 = ModelPackaging(
    id='pack1',
    spec=ModelPackagingSpec(
        integration_name="pi",
        targets=[target_pull, target_push]
    ),
    status=ModelPackagingStatus()
)

pi = PackagingIntegration(id="pi", spec=PackagingIntegrationSpec(
    schema=Schema(
        targets=[
            TargetSchema(
                connection_types=['gcr', 'ecr', 'docker'], name='docker-push', required=True
            ),
            TargetSchema(
                connection_types=['gcr', 'ecr', 'docker'], name='docker-pull', required=False
            )
        ]
    )
))


@dataclass
class I:
    cmd: List[str]
    local: List[Any]
    remote: List[Any]


@dataclass
class E:
    targets: List[PackagerTarget]
    exit_code: int = 0
    exc: BaseException = None
    output_subs: str = ""


@dataclass
class Case:
    input: I
    expected: E


test_cases: List[Case] = [
    Case(  # all manifests are locally stored, default run without options
        input=I(
            cmd=["--pack-id", "pack1"],
            local=[conn1, conn2, pi, pack1], remote=[]),
        expected=E(targets=[])
    ),
    Case(  # no global targets disable
        input=I(
            cmd=["--pack-id", "pack1", "--no-disable-package-targets"],
            local=[conn1, conn2, pi, pack1], remote=[]),
        expected=E(targets=[pi_target_pull, pi_target_push])
    ),
    Case(  # no global targets disable, but specific target is disabled
        input=I(
            cmd=[
                "--pack-id", "pack1", "--no-disable-package-targets",
                "--disable-target", "docker-pull"
            ],
            local=[conn1, conn2, pi, pack1], remote=[]),
        expected=E(targets=[pi_target_push])
    ),
    Case(  # manifests on the server
        input=I(
            cmd=["--pack-id", "pack1"],
            local=[], remote=[conn1, conn2, pi, pack1]),
        expected=E(targets=[])
    ),
    Case(  # some manifests are local stored, some in the server
        input=I(
            cmd=["--pack-id", "pack1"],
            local=[pi, conn2], remote=[conn1, pack1]),
        expected=E(targets=[])
    ),
    Case(  # pack not found
        input=I(
            cmd=["--pack-id", "pack1"],
            local=[], remote=[conn1, conn2, pi]),
        expected=E(targets=[], exit_code=1, exc=WrongHttpStatusCode(404, {"message": f"Not found {pack1.id}"}))
    ),
    Case(  # no global targets disable, conn missed
        input=I(
            cmd=["--pack-id", "pack1", "--no-disable-package-targets"],
            local=[conn1, pi, pack1], remote=[]),
        expected=E(targets=[], exit_code=1, exc=SystemExit(1,),
                   output_subs="\"conn2\" connection of \"docker-push\" target is not found")
    ),
]


@pytest.mark.parametrize("test_case", test_cases)
def test_run_targets(cli_runner: CliRunner, test_case: Case, mocker: MockFixture):

    input_d, expected = test_case.input, test_case.expected
    command_args, local, remote = input_d.cmd, input_d.local, input_d.remote

    # Prepare remote API client mocks
    api_client = Mock()

    def api_retrieve(id_: str):
        for en in remote:
            if id_ == en.id:
                return en
        raise WrongHttpStatusCode(404, {"message": f"Not found {id_}"})

    api_client.get = Mock(side_effect=api_retrieve)
    api_client.get_decrypted = Mock(side_effect=api_retrieve)

    mocker.patch(
        "odahuflow.cli.parsers.local.packaging.PackagingIntegrationClient.construct_from_other",
        new=Mock(return_value=api_client)
    )
    mocker.patch(
        "odahuflow.cli.parsers.local.packaging.ConnectionClient.construct_from_other",
        new=Mock(return_value=api_client)
    )

    # Prepare entities that are stored locally
    with cli_runner.isolated_filesystem():
        with open('manifest.yaml', 'w') as f:
            docs = []
            for en in local:
                d = en.to_dict()
                d["kind"] = en.__class__.__name__
                docs.append(d)
            yaml.safe_dump_all(docs, f)
        if "--manifest-file" not in command_args:
            command_args += ["--manifest-file", "manifest.yaml"]

        # Run
        m: Mock = mocker.patch("odahuflow.cli.parsers.local.packaging.start_package")
        result = cli_runner.invoke(run, command_args, obj=api_client)

        # Assert expectations
        if expected.exit_code == 0:
            assert result.exit_code == 0
            assert result.exception is None

            m.assert_called_once()
            mp_json_actual, _ = m.call_args[0]  # type: K8sPackager, Any

            assert mp_json_actual.targets == expected.targets

        else:
            assert result.exit_code == 1
            assert isinstance(result.exception, type(expected.exc))
            assert str(expected.exc) == str(result.exception)

        if expected.output_subs:
            assert expected.output_subs in result.output



#
#    Copyright 2020 EPAM Systems
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
"""
Local packaging commands for odahuflow cli
"""
import base64
import logging
from typing import List, Dict, Optional

import click
from click import ClickException

from odahuflow.cli.utils import click_utils
from odahuflow.cli.utils.client import pass_obj
from odahuflow.sdk import config
from odahuflow.sdk.clients.api import WrongHttpStatusCode
from odahuflow.sdk.clients.api_aggregated import \
    parse_resources_file, \
    parse_resources_dir, OdahuflowCloudResourceUpdatePair
from odahuflow.sdk.clients.connection import ConnectionClient
from odahuflow.sdk.clients.packaging import ModelPackagingClient
from odahuflow.sdk.clients.packaging_integration import PackagingIntegrationClient
from odahuflow.sdk.local.packaging import start_package, cleanup_packaging_docker_containers
from odahuflow.sdk.models import Connection, K8sPackager, ModelPackaging, PackagerTarget, \
    PackagingIntegration, Target, TargetSchema

LOGGER = logging.getLogger(__name__)


@click.group(name='packaging', cls=click_utils.BetterHelpGroup)
@click.option('--url', help='API server host', default=config.API_URL)
@click.option('--token', help='API server jwt token', default=config.API_TOKEN)
@click.pass_context
def packaging_group(ctx: click.core.Context, url: str, token: str):
    """
    Local packaging process.\n
    Alias for the command is pack.
    """
    ctx.obj = ModelPackagingClient(url, token)


@packaging_group.command('cleanup-containers')
def cleanup_containers():
    """
    \b
    Delete all packaging docker containers.
    \b
    Usage example:
        * odahuflowctl local pack cleanup-containers
    \f
    """
    cleanup_packaging_docker_containers()


def fetch_local_entities(manifest_file: List[str], manifest_dir: List[str]) -> List[OdahuflowCloudResourceUpdatePair]:
    """
    Collect entities from manifest files in local FS and return a result
    Manifests can be collected from file or files inside a directory
    Manifests from different sources are combined together
    """

    entities: List[OdahuflowCloudResourceUpdatePair] = []
    for file_path in manifest_file:
        entities.extend(parse_resources_file(file_path).changes)

    for dir_path in manifest_dir:
        entities.extend(parse_resources_dir(dir_path))

    return entities


def parse_entities(
        entities: List[OdahuflowCloudResourceUpdatePair], pack_id: str
) -> (ModelPackaging, Dict[str, PackagingIntegration], Dict[str, Connection]):
    mp: Optional[ModelPackaging] = None

    packagers: Dict[str, PackagingIntegration] = {}
    connections: Dict[str, Connection] = {}

    for entity in map(lambda x: x.resource, entities):
        if isinstance(entity, PackagingIntegration):
            packagers[entity.id] = entity
        elif isinstance(entity, Connection):
            connections[entity.id] = entity
        elif isinstance(entity, ModelPackaging) and entity.id == pack_id:
            mp = entity

    return mp, packagers, connections


def _decode_connection(connection: Connection) -> None:
    encoding = 'utf-8'
    decode_fields = ['password', 'key_secret', 'key_id', 'public_key']

    for f in decode_fields:
        v: str = getattr(connection.spec, f)
        if not v:
            continue

        try:
            decoded = base64.b64decode(v, validate=True)
        except Exception as e:
            LOGGER.error(f'Unable to decode base64 .spec.{f} of connection {connection.id}')
            raise e

        try:
            decoded_string = decoded.decode(encoding)
        except Exception as e:
            LOGGER.error(f'Unable to decode utf-8 .spec.{f} of connection {connection.id}')
            raise e

        setattr(connection.spec, f, decoded_string)


def get_packager(
        name: str, local: Dict[str, PackagingIntegration], remote_api: PackagingIntegrationClient
) -> PackagingIntegration:
    """
    Fetch Packager entity by name looking in local manifests at first
    and trying to fetch from web api after
    :param name: name of packager
    :param local: entities parsed from local manifests
    :param remote_api: client to fetch packagers from API server
    :return:
    """
    packager = local.get(name)
    if not packager:
        click.echo(
            f'The "{name}" packager is not found in the manifest files. '
            f'Trying to retrieve it from API server'
        )
        packager = remote_api.get(name)
    return packager


def get_packager_targets(
        targets: List[Target], connections: Dict[str, Connection], remote_api: ConnectionClient
) -> List[PackagerTarget]:
    """
    Build targets for calling packager. Fetch and base64 decode connections by names using local manifest and
    ODAHU connections API
    :param targets: Targets from packaging manifest
    :param connections: Connections found in local manifest files
    :param remote_api: ConnectionClient to fetch missing Connections
    """

    packager_targets: List[PackagerTarget] = []

    for t in targets:
        conn = connections.get(t.connection_name)
        if not conn:
            click.echo(
                f'The "{t.connection_name}" connection of "{t.name}" target is not found in the manifest files. '
                f'Trying to retrieve it from API server'
            )
            conn = remote_api.get_decrypted(t.connection_name)

        _decode_connection(conn)

        packager_targets.append(
            PackagerTarget(conn, t.name)
        )

    return packager_targets


def get_packager_target(
        target: Target, connections: Dict[str, Connection], remote_api: ConnectionClient
) -> PackagerTarget:
    """
    Build targets for calling packager. Fetch and base64 decode connections by names using local manifest and
    ODAHU connections API
    :param target: Targets from packaging manifest
    :param connections: Connections found in local manifest files
    :param remote_api: ConnectionClient to fetch missing Connections
    """

    conn = connections.get(target.connection_name)
    if not conn:
        click.echo(
            f'The "{target.connection_name}" connection of "{target.name}" target is not found in the manifest files. '
            f'Trying to retrieve it from API server'
        )
        conn = remote_api.get_decrypted(target.connection_name)

    _decode_connection(conn)

    return PackagerTarget(conn, target.name)


def _deprecation_warning(is_target_disabled: bool):
    if is_target_disabled is None:
        click.echo('[FeatureWarning] Current behavior is to disable all packager targets for a local run. '
                   'In future releases this behavior will be removed '
                   'and all targets from ModelPackaging manifest will be enabled by default. '
                   'Consider using --disable-target option to disable specific targets by name')
    else:
        click.echo('[FeatureWarning] --disable-package-targets/--no-disable-package-targets options '
                   'are deprecated and will be removed in future releases. '
                   'By default all targets will be enabled. '
                   'Consider using --disable-target option to disable specific targets by name')


def validate_targets(
        targets: List[Target], pi: PackagingIntegration, disabled_targets: List[str],
        connections: Dict[str, Connection], remote_api: ConnectionClient
) -> List[PackagerTarget]:

    result: List[PackagerTarget] = []
    errors = []

    targets_in_spec = {t.name for t in targets}
    schema = {t.name: t for t in pi.spec.schema.targets}

    # Set defaults
    for ts in pi.spec.schema.targets:
        if ts.name in targets_in_spec:  # already in spec
            continue
        if ts.name in disabled_targets:  # disabled
            continue
        targets_in_spec.add(ts.name)
        targets.append(Target(ts.default, ts.name, ))
        LOGGER.info(f'{ts.name} target default value is set to ({ts.default})')

    # validate targets
    for t in targets:
        t_schema: TargetSchema = schema.get(t.name)
        if t_schema is None:
            errors.append(f'cannot find "{t.name}" target in "{pi.id}" packaging integration')
            continue

        try:
            pt = get_packager_target(t, connections, remote_api)
        except WrongHttpStatusCode as e:
            if e.status_code == 404:
                errors.append(f'"{t.connection_name}" connection of "{t.name}" target is not found')
                continue
            raise e

        if pt.connection.spec.type not in t_schema.connection_types:
            errors.append(f'"{pt.name}" target has invalid connection type, "{pt.connection.spec.type}", '
                          f'for "{pi.id}" packaging integration the expected are {t_schema.connection_types}')
            continue

        result.append(pt)

    if errors:
        raise ClickException('Next errors were found during .spec.targets validation:\n'
                             + '\n'.join(errors))

    return result


@packaging_group.command()
@click.option('--pack-id', '--id', help='Model packaging ID', required=True)
@click.option('--manifest-file', '-f', type=click.Path(), multiple=True,
              help='Path to an ODAHU-flow manifest file')
@click.option('--manifest-dir', '-d', type=click.Path(), multiple=True,
              help='Path to a directory with ODAHU-flow manifest files')
@click.option('--artifact-path', type=click.Path(),
              help='Path to a training artifact')
@click.option('--artifact-name', '-a', type=str, help='Override artifact name from file')
# TODO Breaking Changes: remove --disable-package-targets/--no-disable-package-targets' options
@click.option('--disable-package-targets/--no-disable-package-targets', 'is_target_disabled',
              default=None, help='Disable all targets in packaging')
@click.option('--disable-target', multiple=True,
              help='Disable target in packaging')
@pass_obj
def run(client: ModelPackagingClient, pack_id: str, manifest_file: List[str], manifest_dir: List[str],
        artifact_path: str, artifact_name: str, is_target_disabled: bool, disable_target: List[str]):
    """
    \b
    Start a packaging process locally.
    \b
    Usage example:
        * odahuflowctl local pack run --id wine
    \f
    """

    _deprecation_warning(is_target_disabled)

    if is_target_disabled is None:  # Backward compatibility
        is_target_disabled = True

    entities: List[OdahuflowCloudResourceUpdatePair] = fetch_local_entities(manifest_file, manifest_dir)

    mp, packagers, connections = parse_entities(entities, pack_id)

    if not mp:
        click.echo(
            f'The "{pack_id}" packaging is not found in the manifest files. '
            f'Trying to retrieve it from API server'
        )
        mp = client.get(pack_id)

    packager = get_packager(
        mp.spec.integration_name, packagers, PackagingIntegrationClient.construct_from_other(client)
    )

    if artifact_name:
        mp.spec.artifact_name = artifact_name
        LOGGER.debug('Override the artifact name')

    if disable_target:
        LOGGER.debug(f'Next targets are disabled: {", ".join(disable_target)}')

    mp_spec_targets = mp.spec.targets if mp.spec.targets is not None else []

    # check for deprecated option
    if is_target_disabled:
        targets = []
    else:
        # Disable targets
        mp_spec_targets = [t for t in mp_spec_targets if t.name not in disable_target]

        # Validate targets
        targets = validate_targets(mp_spec_targets, packager, disable_target,
                                   connections, ConnectionClient.construct_from_other(client))

    k8s_packager = K8sPackager(
        model_packaging=mp,
        packaging_integration=packager,
        targets=targets,
    )

    result = start_package(k8s_packager, artifact_path)

    click.echo('Packager results:')
    for key, value in result.items():
        click.echo(f'* {key} - {value}')



#  Copyright 2020 EPAM Systems
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from unittest.mock import Mock
from click.testing import CliRunner
from pytest_mock import MockFixture
from odahuflow.cli.parsers import security
from odahuflow.sdk.clients.api import APIConnectionException


# TODO: Write a complete set of unit tests for security module

def test_login_invalid_url(cli_runner: CliRunner, mocker: MockFixture):
    """
    Tests fix for issue #257 (https://github.com/odahu/odahu-flow/issues/257)
    If
    :param cli_runner: Click CLI runner fixture
    :param mocker: pytest mocker fixture
    """

    mocker.patch.object(security, 'update_config_file')

    api_client_class_mock: Mock = mocker.patch.object(security.api, 'RemoteAPIClient')
    api_client: Mock = api_client_class_mock.return_value
    api_client.info.side_effect = APIConnectionException

    result = cli_runner.invoke(security.login, ['--url', '--help'])
    assert result.exit_code != 0
    assert isinstance(result.exception, APIConnectionException)



#
#    Copyright 2019 EPAM Systems
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
"""
Security commands for odahuflow cli
"""
import logging

import click
from click import UsageError

from odahuflow.sdk.clients import api
from odahuflow.sdk.config import update_config_file

LOGGER = logging.getLogger(__name__)


def _reset_credentials():
    """
    Clean credentials from config file
    :return:
    """
    update_config_file(API_URL=None,
                       API_TOKEN=None,
                       API_REFRESH_TOKEN=None,
                       API_ACCESS_TOKEN=None,
                       API_ISSUING_URL=None,
                       ODAHUFLOWCTL_OAUTH_CLIENT_ID=None,
                       ODAHUFLOWCTL_OAUTH_CLIENT_SECRET=None,
                       ISSUER_URL=None,
                       MODEL_HOST=None)


@click.command()
@click.option('--url', 'api_host', help='API server host', required=True)
@click.option('--token', help='API server jwt token')
@click.option('--client_id', help='client_id for OAuth2 Client Credentials flow')
@click.option('--client_secret', help='client_secret for OAuth2 Client Credentials flow')
@click.option('--issuer', help='OIDC Issuer URL')
def login(api_host: str, token: str, client_id: str, client_secret: str, issuer: str):
    """
    Authorize on API endpoint.
    Check that credentials is correct and save to the config
    """

    # clean config from previous credentials
    _reset_credentials()

    # update config
    update_config_file(
        API_URL=api_host,
        API_TOKEN=token,
        ODAHUFLOWCTL_OAUTH_CLIENT_ID=client_id,
        ODAHUFLOWCTL_OAUTH_CLIENT_SECRET=client_secret,
        ISSUER_URL=issuer
    )

    # set predicates
    is_token_login = bool(token)
    is_client_cred_login = bool(client_id) or bool(client_secret)
    is_interactive_login = not (is_token_login or is_client_cred_login)

    # validate
    if is_token_login and is_client_cred_login:
        raise UsageError('You should use either --token or --client_id/--client_secret to login. '
                         'Otherwise skipp all options to launch interactive login mode')
    if is_client_cred_login and (not client_id or not client_secret):
        raise UsageError('You must pass both client_id and client_secret to client_credentials login')
    if is_client_cred_login and not issuer:
        raise UsageError('You must provide --issuer parameter to do client_credentials login. '
                         'Or set ISSUER_URL option in config file')

    # login
    api_client = api.RemoteAPIClient(api_host, token, client_id, client_secret,
                                     non_interactive=not is_interactive_login,
                                     issuer_url=issuer)
    try:
        api_client.info()
        print('Success! Credentials have been saved.')
    except api.IncorrectAuthorizationToken as wrong_token:
        LOGGER.error('Wrong authorization token\n%s', wrong_token)
        raise
    except api.APIConnectionException as connection_exc:
        LOGGER.error(f'Failed to connect to API host! '
                     f'Error: {connection_exc}')
        raise


@click.command()
def logout():
    """
    Remove all authorization data from the configuration file
    """
    _reset_credentials()
    click.echo('All authorization credentials have been removed')
